@@ -35,6 +35,10 @@
 #define CHUNK_HORIZONTAL 0
 #define CHUNK_VERTICAL 1
 
+/* private prototypes */
+gboolean gok_chunker_chunk_each_key (GokKeyboard* keyboard, ChunkOrder chunkOrder);
+
+
 /* cell that is currently highlighted */
 static gint m_CellHighlightedRow;
 static gint m_CellHighlightedColumn;
@@ -49,8 +53,8 @@ static GokChunkitem* m_pChunkitemHighlighted;
 static gboolean g_bChunkSelected;
 
 /* type of chunks in use */
-static gint m_ChunkType;
-static gint m_ChunkOrder;
+static ChunkTypes m_ChunkType;
+static ChunkOrder m_ChunkOrder;
 
 /* if set then disabled keys are not included in chunks */
 static gboolean m_bIgnoreDisabledInChunks = TRUE;
@@ -131,6 +135,44 @@ gboolean gok_chunker_chunk_none ()
 }
 
 
+/**
+* gok_chunker_chunk_keys
+* @LeftRight: Direction the keys are ordered. 
+* @TopBottom: Direction the rows are ordered.
+*
+* Breaks all the keyboards into rows and keys.
+*
+* returns: TRUE if everything went OK, FALSE if there was any problem.
+**/
+gboolean gok_chunker_chunk_keys (gboolean TopBottom, gboolean LeftRight)
+{
+	gint orderChunks;
+
+	if (LeftRight == 0)
+	{
+		if (TopBottom == 0)
+		{
+			orderChunks = CHUNKER_TOPTOBOTTOM_LEFTTORIGHT;
+		}
+		else
+		{
+			orderChunks = CHUNKER_BOTTOMTOTOP_LEFTTORIGHT;
+		}
+	}
+	else
+	{
+		if (TopBottom == 0)
+		{
+			orderChunks = CHUNKER_TOPTOBOTTOM_RIGHTTOLEFT;
+		}
+		else
+		{
+			orderChunks = CHUNKER_BOTTOMTOTOP_RIGHTTOLEFT;
+		}
+	}
+	return gok_chunker_chunk_all (CHUNKER_KEYS, orderChunks);
+}
+
 /**
 * gok_chunker_chunk_rows
 * @LeftRight: Direction the keys are ordered. 
@@ -217,7 +259,7 @@ gboolean gok_chunker_chunk_columns (gboolean LeftRight, gboolean TopBottom)
 *
 * returns: TRUE if everything went OK, FALSE if there was any problem.
 **/
-gboolean gok_chunker_chunk_all (gint TypeChunks, gint OrderChunks)
+gboolean gok_chunker_chunk_all (ChunkTypes TypeChunks, ChunkOrder OrderChunks)
 {
 	GokKeyboard* pKeyboard;
 	gboolean returnValue = TRUE;
@@ -254,7 +296,7 @@ void gok_chunker_dump_chunks ()
 	pKeyboard = gok_main_get_first_keyboard();
 	while (pKeyboard != NULL)
 	{
-		printf ("++++++keyboard: %s\n", pKeyboard->Name);
+		gok_log ("++++++keyboard: %s\n", pKeyboard->Name);
 		gok_chunker_dump_chunk (pKeyboard->pChunkFirst);
 
 		pKeyboard = pKeyboard->pKeyboardNext;
@@ -272,19 +314,19 @@ void gok_chunker_dump_chunk (GokChunk* pChunk)
 	GokChunkitem* pChunkitem;
 	while (pChunk != NULL)
 	{
-		printf ("----chunk:%d\n", pChunk->ChunkId);
+		gok_log ("----chunk:%d\n", pChunk->ChunkId);
 		pChunkitem = pChunk->pChunkitem;
 		while (pChunkitem != NULL)
 		{
-			printf ("%s\n", gok_key_get_label (pChunkitem->pKey));
+			gok_log ("%s\n", gok_key_get_label (pChunkitem->pKey));
 			pChunkitem = pChunkitem->pItemNext;
 		}
 	
 		if (pChunk->pChunkChild != NULL)
 		{
-			printf (">>>>>Child<<<<<<<\n");
+			gok_log (">>>>>Child<<<<<<<\n");
 			gok_chunker_dump_chunk (pChunk->pChunkChild);
-			printf (">>>>>Child END<<<<<<<\n");
+			gok_log (">>>>>Child END<<<<<<<\n");
 		}
 		
 		pChunk = pChunk->pChunkNext;
@@ -321,6 +363,11 @@ gboolean gok_chunker_chunk (GokKeyboard* pKeyboard)
 			/* keyboard is not broken into chunks (e.g. direct selection) */
 			break;
 
+		case CHUNKER_KEYS:
+			/* todo: implement ordering */
+			returnValue = gok_chunker_chunk_each_key (pKeyboard, (ChunkOrder)m_ChunkOrder);
+			break;
+
 		case CHUNKER_ROWS:
 			if ((m_ChunkOrder == CHUNKER_TOPTOBOTTOM_LEFTTORIGHT) ||
 				(m_ChunkOrder == CHUNKER_TOPTOBOTTOM_RIGHTTOLEFT))
@@ -389,6 +436,173 @@ gboolean gok_chunker_chunk (GokKeyboard* pKeyboard)
 	return returnValue;
 }
 
+/* private helper */
+GokChunk*
+gok_chunker_get_bestchunk (GokChunk* chunk, ChunkOrder order) 
+{
+	GokChunk* bestChunk;
+	GokKey* bestKey;
+	GokKey* key;
+	gboolean better;
+	
+	gok_log_enter();
+	bestChunk = chunk;
+	bestKey = bestChunk->pChunkitem->pKey;
+	
+	while ((chunk = chunk->pChunkNext) != NULL) {
+		better = FALSE;
+		key = chunk->pChunkitem->pKey;
+		switch (order) {
+			case CHUNKER_TOPTOBOTTOM_LEFTTORIGHT:
+				if ((key->Top < bestKey->Top) ||
+					((key->Top == bestKey->Top) && 
+					 (key->Left < bestKey->Left))) {
+						 better = TRUE;
+				}
+				break;	
+			case CHUNKER_TOPTOBOTTOM_RIGHTTOLEFT:
+				if ((key->Top < bestKey->Top) ||
+					((key->Top == bestKey->Top) && 
+					 (key->Left > bestKey->Left))) {
+						 better = TRUE;
+				}
+				break;	
+			case CHUNKER_BOTTOMTOTOP_LEFTTORIGHT:
+				if ((key->Top > bestKey->Top) ||
+					((key->Top == bestKey->Top) && 
+					 (key->Left < bestKey->Left))) {
+						 better = TRUE;
+				}
+				break;	
+			case CHUNKER_BOTTOMTOTOP_RIGHTTOLEFT:
+				if ((key->Top > bestKey->Top) ||
+					((key->Top == bestKey->Top) && 
+					 (key->Left > bestKey->Left))) {
+						 better = TRUE;
+				}
+				break;	
+		}
+		if (better == TRUE) {
+			bestChunk = chunk;
+			bestKey = bestChunk->pChunkitem->pKey;
+		}
+	}
+	gok_log_leave();
+	return bestChunk;
+}
+
+/* private helper */
+void
+gok_chunker_chunk_move_before (GokChunk* gc1, GokChunk* gc2) 
+{
+	gok_log("inserting [%s] before [%s]",
+		gok_key_get_label(gc2->pChunkitem->pKey),
+		gok_key_get_label(gc1->pChunkitem->pKey));
+
+	/* assume gc2 is later than gc1 in the list */
+	/* remove gc2*/
+	if (gc2->pChunkPrevious != NULL) { 
+		gc2->pChunkPrevious->pChunkNext = gc2->pChunkNext;
+	}
+	if (gc2->pChunkNext != NULL) {
+		gc2->pChunkNext->pChunkPrevious = gc2->pChunkPrevious;
+	}
+	/* insert gc2 before gc1 */
+	if (gc1->pChunkPrevious != NULL) { gc1->pChunkPrevious->pChunkNext = gc2; }
+	gc2->pChunkPrevious = gc1->pChunkPrevious;
+	gc2->pChunkNext = gc1;
+	gc1->pChunkPrevious = gc2;
+}
+
+		
+/* private helper */
+GokChunk*
+gok_chunker_sort_key_chunks (GokChunk* headChunk, ChunkOrder chunkOrder)
+{
+	GokChunk* chunk;
+	GokChunk* bestChunk;
+	GokChunk* newHeadChunk;
+	gok_log_enter();
+	chunk = headChunk;
+	newHeadChunk = NULL;
+	while (chunk != NULL) {
+		gok_log ("chunk [%s]",gok_key_get_label(chunk->pChunkitem->pKey));
+		bestChunk = gok_chunker_get_bestchunk(chunk, chunkOrder);
+		if (newHeadChunk == NULL) { newHeadChunk = bestChunk; /* done once */}
+		if (bestChunk != chunk) {
+			gok_chunker_chunk_move_before (chunk, bestChunk);
+		}
+		else {
+			chunk = chunk->pChunkNext;
+		}
+	}
+	gok_log_leave();
+	return newHeadChunk;
+}
+
+/**
+* gok_chunker_chunk_each_key
+* @keyboard: Pointer to the keyboard that will be broken ginto chunks of rows.
+* @chunkOrder: Defines the row order (left to right or right to left). See enum ChunkOrder.
+*
+* Chunk entire keyboard into single key chunks.
+*
+* returns: TRUE if the keyboard was broken into keys, FALSE if not
+**/
+gboolean
+gok_chunker_chunk_each_key (GokKeyboard* keyboard, ChunkOrder chunkOrder)
+{
+	GokKey* key;
+	GokChunk* chunkNew;
+	GokChunk* chunkLast;
+	GokChunk* chunkFirst;
+	GokChunkitem* chunkitemNew;
+	gint i;
+	
+	gok_log_enter();
+	
+	chunkFirst = NULL;
+	chunkLast = NULL;
+	
+	/* iterate through all the keys on the keyboard */
+	key = keyboard->pKeyFirst;
+	i = 0;
+	while (key != NULL)
+	{
+		g_assert (key->pButton != NULL);
+
+		/* ignore keys that are disabled */
+		if ((m_bIgnoreDisabledInChunks == TRUE) &&
+			(strcmp (gtk_widget_get_name (key->pButton), "StyleButtonDisabled") == 0))
+		{
+			key = key->pKeyNext;
+			continue;
+		}
+		/* create a new chunk for each key */
+		chunkNew = (GokChunk*) g_malloc (sizeof (GokChunk));
+		gok_chunker_init_chunk (chunkNew);
+		chunkNew->ChunkId = i++;
+		chunkNew->pChunkPrevious = chunkLast;
+		if (chunkFirst == NULL) {
+			chunkFirst = chunkNew;
+		}
+		else {
+			chunkLast->pChunkNext = chunkNew;
+		}
+		/* create a chunkitem for each chunk */
+		chunkitemNew = (GokChunkitem*) g_malloc (sizeof (GokChunkitem));
+		gok_chunker_init_chunkitem (chunkitemNew);
+		chunkitemNew->pKey = key;
+		chunkNew->pChunkitem = chunkitemNew;
+
+		chunkLast = chunkNew;
+		key = key->pKeyNext;
+	}
+	keyboard->pChunkFirst = gok_chunker_sort_key_chunks (chunkFirst, chunkOrder);
+	gok_log_leave();
+	return TRUE;
+}
+
 /**
 * gok_chunker_chunk_rows_ttb
 * @pKeyboard: Pointer to the keyboard that will be broken ginto chunks of rows.
@@ -1004,7 +1218,7 @@ gboolean gok_chunker_chunk_recursive (GokKeyboard* pKeyboard, gint ChunkOrder, g
 * Returns: A pointer to the first chunk in the list of chunks. Returns
 * NULL if the given chunk can't be divided into chunks.
 **/
-GokChunk* gok_chunker_chunk_group (GokChunk* pChunk, gint NumberGroups, gint Order, gboolean bCanReorder)
+GokChunk* gok_chunker_chunk_group (GokChunk* pChunk, gint NumberGroups, ChunkOrder Order, gboolean bCanReorder)
 {
 	GokChunk* pFirstChunk;
 	GokChunk* pChunkLoop;
@@ -1787,6 +2001,7 @@ void gok_chunker_delete_chunks (GokChunk* pChunk, gboolean bAlsoNext)
 	GokChunkitem* pChunkitem;
 	GokChunkitem* pChunkitemTemp;
 
+	gok_log_enter();
 	while (pChunk != NULL)
 	{
 		/* delete all the chunk items */
@@ -1822,6 +2037,7 @@ void gok_chunker_delete_chunks (GokChunk* pChunk, gboolean bAlsoNext)
 			break;
 		}
 	}
+	gok_log_leave();
 }
 
 /**
@@ -2158,6 +2374,7 @@ void gok_chunker_highlight_chunk (GokChunk* pChunk)
 	GokChunkitem* pChunkitem;
 
 	g_assert (pChunk != NULL);
+	gok_log_enter();
 
 	pChunkitem = pChunk->pChunkitem;
 	while (pChunkitem != NULL)
@@ -2173,6 +2390,7 @@ void gok_chunker_highlight_chunk (GokChunk* pChunk)
 
 	/* no key selected if a chunk is  highlighted */
 	gok_feedback_set_selected_key (NULL);
+	gok_log_leave();
 }
 
 /**
@@ -2285,8 +2503,11 @@ void gok_chunker_unhighlight_all_keys ()
 **/
 void gok_chunker_highlight_first_key ()
 {
-	g_assert (m_pChunkHighlighted != NULL);
-
+	if (m_pChunkHighlighted == NULL) {
+		gok_log_x( "aborting - no highlighted chunk to scan");
+		return;
+	}
+	
 	m_pChunkitemHighlighted = m_pChunkHighlighted->pChunkitem;
 	g_assert (m_pChunkitemHighlighted != NULL);
 	g_assert (m_pChunkitemHighlighted->pKey != NULL);
@@ -2662,7 +2883,7 @@ void gok_chunker_keyhighlight()
 gboolean gok_chunker_wraptofirstchunk ()
 {
 	GokChunk* pChunkFirst;
-	
+
 	if (m_pChunkHighlighted == NULL)
 	{
 		gok_log_x ("Warning: m_pChunkHighlighted is NULL!\n");
