@@ -39,21 +39,13 @@
 #include "support.h"
 #include "gok-log.h"
 #include "main.h"
+#include "gok-gconf-keys.h"
 
 /* pointer to the settings dialog window */
 static GtkWidget* m_pWindowSettings;
 
 static gboolean is_locked;
 
-void
-gok_settingsdialog_lock (void)
-{
-	is_locked = TRUE;
-}
-
-
-/* Private functions */
-
 /**
 * gok_settingsdialog_open
 * 
@@ -81,6 +73,8 @@ gboolean gok_settingsdialog_open (gboolean bShow)
 		return FALSE;
 	}
 
+	is_locked = FALSE;  /* don't read gconf here */
+
 	/* initialize all the pages */
 	gok_settings_page_keysizespace_initialize (m_pWindowSettings);
 	gok_page_accessmethod_initialize (m_pWindowSettings);
@@ -200,10 +194,9 @@ gboolean gok_settingsdialog_open (gboolean bShow)
                       NULL);
 
 	
-	/* show/hide the settings dialog */
 	if (bShow == TRUE)
 	{
-		gtk_widget_show (m_pWindowSettings);
+		gok_settingsdialog_show ();
 	}
 
 	return TRUE;
@@ -220,19 +213,30 @@ gboolean gok_settingsdialog_show()
 {
 	gboolean bFirstShowing;
 	GtkWidget *label;
+	GtkWidget *notebook;
+	gboolean was_locked;
 
 	g_assert (m_pWindowSettings != NULL);
 
-	bFirstShowing = (GDK_IS_DRAWABLE(m_pWindowSettings->window) == FALSE) ? TRUE : FALSE;
-
-	/* if we've "locked" certain prefs, de-sensitize the relevant page tabs */
-	if (is_locked) {
-		label = lookup_widget (m_pWindowSettings, "label2");
-		gtk_widget_set_sensitive (label, FALSE);
-		label = lookup_widget (m_pWindowSettings, "label4");
-		gtk_widget_set_sensitive (label, FALSE);
+	was_locked = is_locked;
+	gok_gconf_get_bool (gconf_client_get_default(),
+		GOK_GCONF_PREFS_LOCKED, &is_locked);
+	
+	if (was_locked != is_locked) {
+		notebook = lookup_widget (m_pWindowSettings, "notebook1");
+		/* FIXME: magic #'s used - easily fixed later when we libglade-ify */
+		if (!is_locked) {
+			gtk_widget_show(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1));
+			gtk_widget_show(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 3));
+		}
+		else {
+			gtk_widget_hide(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1));
+			gtk_widget_hide(gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 3));
+		}
 	}
 
+	bFirstShowing = (GDK_IS_DRAWABLE(m_pWindowSettings->window) == FALSE) ? TRUE : FALSE;
+
 	gtk_widget_show (m_pWindowSettings);
 
 	/* if this is the first time we're showing the dialog then redisplay
@@ -276,10 +280,16 @@ GtkWidget* gok_settingsdialog_get_window ()
 **/
 void gok_settingsdialog_close()
 {
-	g_assert (m_pWindowSettings != NULL);
-	gtk_widget_destroy (m_pWindowSettings);
+	if (m_pWindowSettings != NULL)
+	{
+		gtk_widget_destroy (m_pWindowSettings);
 	
-	gok_page_accessmethod_close();
+		gok_page_accessmethod_close();
+	}
+	else
+	{
+		g_warning("The settings dialog doesn't exist in close function!");
+	}
 }
 
 /**
@@ -494,4 +504,3 @@ gint gok_settingsdialog_sort (gconstpointer pItem1, gconstpointer pItem2)
 {
 	return strcmp (pItem1, pItem2);
 }
-
