@@ -42,6 +42,7 @@
 #include "gok-data.h"
 #include "gok-modifier-keymasks.h"
 #include "gok-gconf-keys.h"
+#include "gok-word-complete.h"
 
 /*
  * NOTE: we can tweak the efficiency/accuracy of the
@@ -103,7 +104,12 @@ static void gok_spy_window_activate_listener (const AccessibleEvent *event,
 static void gok_spy_window_deactivate_listener (const AccessibleEvent *event,
                                                 void *user_data);
 												
+static void gok_spy_object_state_listener (const AccessibleEvent *event,
+                                                void *user_data);
+
 static gboolean gok_spy_is_visible (Accessible *accessible);
+
+static void gok_spy_resolve_namesakes (AccessibleNode *node);
 												
 
 /*
@@ -114,6 +120,7 @@ static GQueue* eventQueue;
 static void* m_ptheAppChangeListener;
 static void* m_ptheWindowChangeListener;
 static void* m_ptheMouseButtonListener;
+static void* m_ptheStateChangeListener;
 
 /* pointer to the current window */
 static Accessible* m_ptheWindowAccessible;
@@ -127,6 +134,7 @@ static AccessibleDeviceListener* mouseListener;
 static AccessibleEventListener* modifierListener;
 static AccessibleEventListener* windowActivateListener;
 static AccessibleEventListener* windowDeactivateListener;
+static AccessibleEventListener* objectStateListener;
 static gboolean m_gokSpyOpen = FALSE;
 static gint explicitrefs;
 static gint explicitunrefs;
@@ -288,6 +296,7 @@ void gok_spy_open()
 		m_ptheMouseButtonListener =NULL;
 		m_ptheAppChangeListener = NULL;
 		m_ptheWindowChangeListener = NULL;
+		m_ptheStateChangeListener = NULL;
 		m_ptheWindowAccessible = NULL;
 		m_ptheAccessibleWithText = NULL;
 
@@ -303,6 +312,8 @@ void gok_spy_open()
 			gok_spy_window_activate_listener, NULL);
 		windowDeactivateListener = SPI_createAccessibleEventListener (
 			gok_spy_window_deactivate_listener, NULL);
+		objectStateListener = SPI_createAccessibleEventListener (
+			gok_spy_object_state_listener, NULL);
 
 		success = SPI_registerGlobalEventListener (focusListener, "focus:"); 
 		gok_log ("SPI_registerGlobalEventListener for focus events returned: %d" , success);
@@ -325,8 +336,8 @@ void gok_spy_open()
 		gok_log ("SPI_registerGlobalEventListener for window:deactivate returned: %d", success);
 
 		success = SPI_registerGlobalEventListener (
-			windowDeactivateListener, "window:destroy");
-		gok_log ("SPI_registerGlobalEventListener for window:destroy returned: %d", success);
+			objectStateListener, "object:state-changed");
+		gok_log ("SPI_registerGlobalEventListener for object:state-changed returned: %d", success);
 
 	}
 	gok_log_leave ();
@@ -386,6 +397,13 @@ void gok_spy_stop (void)
 		AccessibleEventListener_unref(windowDeactivateListener );
 		windowDeactivateListener = NULL;
 	}
+	if (objectStateListener != NULL)
+	{
+		result= SPI_deregisterGlobalEventListenerAll (objectStateListener);
+		gok_log("deregistering object state change listener returned: %d",result);
+		AccessibleEventListener_unref(objectStateListener );
+		objectStateListener = NULL;
+	}
 }
 	
 /** 
@@ -435,7 +453,7 @@ void gok_spy_close(void)
 * 
 * @callback: the listener to register
 **/ 
-void gok_spy_register_appchangelistener(AppChangeListener* callback)
+void gok_spy_register_appchangelistener(AccessibleChangeListener* callback)
 {
 	gok_log_x("this function deprecated, use gok_spy_register_windowchangelistener instead");
 	/* m_ptheAppChangeListener = callback; */
@@ -446,7 +464,7 @@ void gok_spy_register_appchangelistener(AppChangeListener* callback)
 * 
 * @callback: the listener to deregister
 **/ 
-void gok_spy_deregister_appchangelistener(AppChangeListener* callback)
+void gok_spy_deregister_appchangelistener(AccessibleChangeListener* callback)
 {
 	gok_log_x("this function deprecated, use gok_spy_deregister_windowchangelistener instead");
 	/* m_ptheAppChangeListener = NULL; */
@@ -457,7 +475,7 @@ void gok_spy_deregister_appchangelistener(AppChangeListener* callback)
 * 
 * @callback: the listener to register
 **/ 
-void gok_spy_register_windowchangelistener(AppChangeListener* callback)
+void gok_spy_register_windowchangelistener(AccessibleChangeListener* callback)
 {
 	m_ptheWindowChangeListener = callback;
 }
@@ -467,11 +485,37 @@ void gok_spy_register_windowchangelistener(AppChangeListener* callback)
 * 
 * @callback: the listener to deregister
 **/ 
-void gok_spy_deregister_windowchangelistener(AppChangeListener* callback)
+void gok_spy_deregister_windowchangelistener(AccessibleChangeListener* callback)
 {
 	m_ptheWindowChangeListener = NULL;
 }
 
+
+/** 
+* gok_spy_register_objectstatelistener
+* 
+* Will only notify if the state change is in the correct context. The context
+* is this: the state change must be happening within the currently focused
+* application, otherwise the callback is not notified.
+*
+* @callback: the listener to register
+**/ 
+void gok_spy_register_objectstatelistener(AccessibleChangeListener* callback)
+{
+	m_ptheStateChangeListener = callback;
+}
+
+/** 
+* gok_spy_deregister_objectstatelistener
+* 
+* @callback: the listener to deregister
+**/ 
+void gok_spy_deregister_objectstatelistener(AccessibleChangeListener* callback)
+{
+	m_ptheStateChangeListener = NULL;
+}
+
+
 /** 
 * gok_spy_register_mousebuttonlistener
 * 
@@ -646,14 +690,21 @@ gboolean gok_spy_check_queues(void)
 			gok_log("GOKSPY_FOCUS_EVENT");
 			gok_spy_process_accessible(en->paccessible);
 			break;
+		case GOKSPY_STATE_EVENT:
+			gok_log("GOKSPY_STATE_EVENT");
+			/* gok_spy_accessible_ref (en->paccessible); */
+			((AccessibleChangeListenerCB)(m_ptheStateChangeListener))(en->paccessible);
+			break;
 		case GOKSPY_WINDOW_ACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_ACTIVATE_EVENT");
+			/* reset word completion, if it's being used */
+			gok_wordcomplete_reset (gok_wordcomplete_get_default ());
 			/* FIXME !! the ref needs to happen in the original event loop */
 			if (en->paccessible != m_ptheWindowAccessible) {
 				gok_spy_accessible_unref (m_ptheWindowAccessible);
 				m_ptheWindowAccessible = en->paccessible;
 				gok_spy_accessible_ref (m_ptheWindowAccessible);
-				((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
+				((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			}
 			break;
 		case GOKSPY_DEFUNCT_EVENT:
@@ -664,7 +715,7 @@ gboolean gok_spy_check_queues(void)
 			gok_spy_accessible_unref (m_ptheAccessibleWithText);
 			m_ptheWindowAccessible = NULL;
 			m_ptheAccessibleWithText = NULL;
-			((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
+			((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			break;
 		case GOKSPY_WINDOW_DEACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_DEACTIVATE_EVENT");
@@ -682,7 +733,7 @@ gboolean gok_spy_check_queues(void)
 				gok_spy_accessible_unref (m_ptheAccessibleWithText);
 				m_ptheWindowAccessible = NULL;
 				m_ptheAccessibleWithText = NULL;
-				((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
+				((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			}
 			break;
 		default:
@@ -804,8 +855,8 @@ gboolean gok_spy_worth_searching (Accessible* accessible, AccessibleRole role)
  * (pass in zero).
  * @found: out variable to hold the success boolean.
  * 
- * If anything under this accessible* has the specified role, then
- * returns true.
+ * If anything under this accessible* has the specified role (constrained by
+ * depth), then returns true.
  *
  * Returns: true if a child exists, false otherwise.
  */
@@ -1049,10 +1100,107 @@ AccessibleNode* gok_spy_create_list (Accessible *anAccessible,
 
 	gok_spy_find_children (&nodeRoot, type, anAccessibleRole);
 
+	gok_spy_resolve_namesakes (nodeRoot.pnext);
 	gok_log_leave();
 	return nodeRoot.pnext;
 }
 
+/* return a node (if exists) from the list with the given name */
+static AccessibleNode*
+gok_spy_node_get_node_named (AccessibleNode *node, gchar *name)
+{
+	while (node != NULL) {
+		if (strcmp (name, node->pname) == 0) {
+			return node;
+		}
+		node = node->pnext;
+	}
+	return NULL;
+}
+
+/* modify the name for the given node so that it is distinct from namesakes */
+static void
+gok_spy_distinguish_node_name (AccessibleNode *node)
+{
+	AccessibleRelation **relations;
+	Accessible *labelAccessible = NULL;
+	AccessibleRelationType type = -1;
+	char *s;
+	gboolean distinguished;
+	gint i,j, num_labels;
+	i = 0;
+	j = 0;
+	num_labels = 0;
+	distinguished = FALSE;
+	
+	/* find the label for the given node and prepend it to the name */
+	relations = Accessible_getRelationSet(node->paccessible);
+	if (relations != NULL) {
+	 while (relations[i]) {
+		if (!distinguished) {
+			type = AccessibleRelation_getRelationType(relations[i]);
+			if (type == SPI_RELATION_LABELED_BY) {
+				num_labels= AccessibleRelation_getNTargets(relations[i]);
+				for (j=0; j < num_labels; j++) {
+					labelAccessible = 
+						AccessibleRelation_getTarget(relations[i], j);
+					if (labelAccessible != NULL) {
+						s = Accessible_getName (labelAccessible);
+						if (s != NULL) {
+							if ( strlen (s) != 0 )  {
+								gchar buffer[64];
+								strncpy (buffer, node->pname, 63);
+								g_free(node->pname); 
+								node->pname = g_strconcat (s, " ", buffer,NULL);
+								SPI_freeString (s);
+								Accessible_unref (labelAccessible);
+								distinguished = TRUE;
+								break; /* from for loop*/
+							}
+							SPI_freeString(s);
+						}
+						Accessible_unref (labelAccessible);
+					}
+				}
+				/* don't break from while loop (need to unref) */
+			}
+		}
+		AccessibleRelation_unref(relations[i++]);
+	 }
+	}
+}
+
+/**
+ * gok_spy_resolve_namesakes:
+ * @node the start node (in the node list)
+ *
+ * Disambiguate nodes in the given list that have the same name.
+ */
+static void
+gok_spy_resolve_namesakes (AccessibleNode *node)
+{
+	gchar name[64]; /* names longer thank this will not be resolved */
+	AccessibleNode *namesake;
+	gboolean distinguished_source = FALSE;
+	
+	/* for every node in the given list... */
+	while (node != NULL) {
+		namesake = node;
+		strncpy (name, node->pname, 63);
+		/* for every node with the same name as the current node... */
+		while ((namesake = gok_spy_node_get_node_named ( namesake->pnext, name )) 
+				!= NULL) {
+			if (!distinguished_source) {
+				distinguished_source = TRUE;
+				gok_spy_distinguish_node_name ( node );
+			}
+			gok_spy_distinguish_node_name ( namesake );
+		}
+		distinguished_source = FALSE;	
+		node = node->pnext;
+	}
+}
+
 static gboolean
 gok_spy_add_node (AccessibleNode *prev, Accessible *accessible, gint i, char *name)
 {
@@ -1192,19 +1340,23 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 		}
 		
 		if (pName == NULL) {
+			gok_log("still no name, so looking at description...");
 			/* one last try, we pull chars from description */
 			s = Accessible_getDescription (pAccessible);
 			if (s != NULL) {
-				gint len = strlen (pName);
+				gint len = strlen (s);
 				if ( len != 0 )  {
 					if (len > 18) 
 					{
-						gchar *tmp = g_strndup (s, 15);
+						gchar *tmp = NULL;
+						gok_log("shortening description");
+						tmp = g_strndup (s, 15);
 						pName = g_strconcat (tmp, "...", NULL);
 						g_free (tmp);
 					}
 					else 
 					{
+						gok_log("using description");
 						pName = g_strndup (s, len);
 					}
 				}
@@ -1212,6 +1364,7 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 			}
 		}
 		if (pName != NULL) {
+			gok_log("we have a name: [%s]",pName);
 			/* create a new node */
 			gok_spy_add_node (pNode, pAccessible, -1, pName);
 		}
@@ -1233,6 +1386,9 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 			AccessibleText *text = Accessible_getText (pAccessible);
 			gint j = 0, len;
 			gchar *word, *gs = NULL;
+
+			gok_log("still no name, but there is text...");
+
 			if (text) {
 				long int start, end = 0;
 				do {
@@ -1274,6 +1430,7 @@ AccessibleNode* gok_spy_append_node (AccessibleNode* pNode,
 				}
 			}
 			if (pName != NULL) {
+				gok_log ("we have a name: [%s]",pName);
 				gok_spy_add_node (pNode, pAccessible, -1, pName);
 			}
 		}
@@ -1364,6 +1521,24 @@ gok_spy_window_deactivate_listener(const AccessibleEvent *event,
 	gok_log_leave ();
 }
 
+
+static void
+gok_spy_object_state_listener(const AccessibleEvent *event,
+                                 void *user_data)
+{
+	EventNode *en;
+	gok_log_enter ();
+
+	gok_spy_accessible_ref (event->source);
+	en = (EventNode*)g_malloc (sizeof (EventNode));
+	en->paccessible = event->source;
+	en->type = GOKSPY_STATE_EVENT;
+	g_queue_push_tail (eventQueue, en);
+	gok_spy_add_idle_handler ();
+
+	gok_log_leave ();
+}
+
 /** 
 * gok_spy_process_accessible 
 * 
@@ -1445,7 +1620,7 @@ void gok_spy_process_accessible (Accessible* accessible)
 		gok_log("different.");
 		gok_spy_accessible_unref (m_ptheWindowAccessible);
 		m_ptheWindowAccessible = pnewWindowAccessible;
-		((AppChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible); 
+		((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible); 
 	}
 	else
 	{
