@@ -83,6 +83,7 @@
 
 #define KEYBOARD_ACCESSIBILITY_STICKY_KEYS_KEY "/desktop/gnome/accessibility/keyboard/stickykeys_enable"
 #define GCONF_ACCESSIBILITY_KEY "/desktop/gnome/interface/accessibility"
+#define KEYBOARD_ACCESSIBILITY_ENABLE_KEY "/desktop/gnome/accessibility/keyboard/enable"
 
 /* pointer to the first keyboard in the list of keyboards */
 static GokKeyboard* m_pKeyboardFirst = NULL;
@@ -694,23 +695,6 @@ gok_main_object_state_listener (Accessible* pAccessible)
 			pK = pK->pKeyNext;
 		}
 	}
-	/* notes:
-	   we now check to see if the current keyboard requires updating, currently
-	   we simply call gok_main_ds, but if this is too slow we can create a 
-	   method that updates the specific key in question 
-	*/
-	
-/*	AccessibleNode* pAN;
-	if (pKB != NULL) {
-		pAN = pKB->pNodeAccessible;
-		while (pAN != NULL) {
-			if (pAN->paccessible == pAccessible) {
-				gok_main_ds (pKB);
-				break;
-			}
-			pAN = pAN->pnext;
-		}
-	}*/
 }	
 
 
@@ -942,9 +926,8 @@ gboolean gok_main_display_scan_reset ()
 * gok_main_display_scan_previous_premade
 * 
 * Display and scan the first keyboard pulled from the branch-back-stack
-* that is premade. Delete any dynamic keyboards that are on the stack before
-* the first premade keyboard.
-* Don't push the current keyboard on the branch-back-stack.
+* that is premade. Delete any dynamic keyboards that are on the stack 'along
+* the way'. Don't push the current keyboard on the branch-back-stack.
 *
 * returns: TRUE if the previous keyboard was displayed, FALSE if not.
 */
@@ -964,24 +947,15 @@ gboolean gok_main_display_scan_previous_premade ()
 		return gok_main_display_scan_previous();
 	}
 
-	/* check this flag before branching back */
-/*
-	if (gok_data_get_dynamic_branchback() == FALSE)
-	{
-		return;
-	}
-*/
-	
 	/* initialize this array */
 	for (index = 0; index < MAX_DELETABLE_KEYBOARDS; index++)
 	{
 		arrayKeyboardsToDelete[index] = NULL;
 	}
-
+	
 	if (gok_branchbackstack_is_empty() == TRUE)
 	{
 		/* no keyboard on the branch back stack */
-		gok_log("no dynamic keyboards to remove during branch");
 		gok_log_leave();
 		return FALSE;
 	}
@@ -1052,8 +1026,7 @@ gboolean gok_main_display_scan_previous ()
 	bReturnCode = TRUE;
 	pKeyboardToDelete = NULL;
 
-	if (gok_branchbackstack_is_empty() == TRUE)
-	{
+	if (gok_branchbackstack_is_empty() == TRUE) {
 		/* no keyboard on the branch back stack */
 		gok_log_leave();
 		return FALSE;
@@ -1073,14 +1046,11 @@ gboolean gok_main_display_scan_previous ()
 	{
 		/* delete the dynamically created keyboard */
 		/* make sure we branched back before deleting it */
-		if (bReturnCode == TRUE)
-		{
-			if (m_pCurrentKeyboard != pKeyboardToDelete) /* hack to fix potentially nastiness */
-			{
+		if (bReturnCode == TRUE) {
+			if (m_pCurrentKeyboard != pKeyboardToDelete) {
 				gok_keyboard_delete (pKeyboardToDelete,FALSE);
 			}
-			else
-			{
+			else {
 				gok_log_x("an attempt was made to delete the currently displayed keyboard!");
 			}
 		}
@@ -1283,6 +1253,7 @@ GtkWidget* gok_main_create_window (gboolean is_dock)
 				"allow_grow", TRUE,
 				"allow_shrink", TRUE,
 				"border_width", 0,
+				"accept_focus", FALSE,
 				NULL),
 			     "signal::realize", on_window1_realize, is_dock,
 			     "signal::destroy", on_window1_destroy, NULL,
@@ -1380,7 +1351,7 @@ void gok_main_read_keyboards ()
     GError *gconf_err = NULL;
     gchar *directory_name = NULL;
     gboolean gconf_keyboard_directory_error = FALSE;
-
+	
     gconf_client = gconf_client_get_default ();
     
     directory_name = gconf_client_get_string (gconf_client,
@@ -1416,9 +1387,10 @@ void gok_main_read_keyboards ()
 
 /**
 * gok_main_read_keyboards_from_dir:
-* @directory: The name of the directory to read the keyboard files from.
+* @directory: The name of the keyboard root directory location.
 * 
-* Reads all the keyboard files from the given directory.
+* Reads all the keyboard files from the given directory or a subdirectory that
+* is a closer match based on user's language if it exists.
 *
 * Returns: A pointer to the first keyboard, NULL if no keyboards could be read.
 */
@@ -1431,16 +1403,40 @@ GokKeyboard* gok_main_read_keyboards_from_dir (const char *directory)
 	struct dirent* pDirectoryEntry;
 	gchar* filename;
 	gchar* complete_path;
+	GList* langlist;
+	gchar dirtemp[1024]; /* no reliable way to avoid magic # here */
 
 	gok_log_enter();
-	gok_log ("directory: %s", directory);
 	pKeyboardFirst = NULL;
 	pKeyboard = NULL;
 	pKeyboardPrevious = NULL;
+	langlist == NULL;
+	pDirectoryKeyboards = NULL;
+
+	/* find the locale specific keyboards */
+	langlist = (GList*) gnome_i18n_get_language_list(NULL);
+	if (langlist != NULL) {
+		gint i = 0;
+		guint l = g_list_length(langlist);
+		while ((pDirectoryKeyboards == NULL) && (i < l)) {
+			strncpy(dirtemp, directory, 950);
+			strncat(dirtemp, G_DIR_SEPARATOR_S, 1);
+			strncat(dirtemp, g_list_nth_data(langlist, i),70);
+			pDirectoryKeyboards = opendir (dirtemp);
+			i++;
+		}
+		g_list_free(langlist);
+	}
+
+	if (pDirectoryKeyboards == NULL) { /* no locale specific directory found */
+		pDirectoryKeyboards = opendir (directory); /* use default */
+	}
+	else {
+		directory = dirtemp; /* used later */
+	}
+		
 
 	/* get a list of all the files from the given directory name */
-	/* first, open the keyboard directory */
-	pDirectoryKeyboards = opendir (directory);
 	if (pDirectoryKeyboards == NULL)
 	{
 		gok_log_x ("Can't open keyboard directory in gok_main_read_files!");
@@ -2133,7 +2129,11 @@ gok_main_check_sticky_keys (GtkWidget *widget)
 			/* Turn on sticky keys and warn user! */
 #if USE_GCONF
 			if (1) {
+				gboolean keyboard_access;
 				GConfClient *gconf_client = gconf_client_get_default ();
+				gok_gconf_get_bool (gconf_client, KEYBOARD_ACCESSIBILITY_ENABLE_KEY, &keyboard_access);
+				if (!keyboard_access)
+					gok_gconf_set_bool (gconf_client, KEYBOARD_ACCESSIBILITY_ENABLE_KEY, TRUE); /* set keyboard accessibility key before setting sticky keys */
 				gok_gconf_set_bool (gconf_client, KEYBOARD_ACCESSIBILITY_STICKY_KEYS_KEY, 
 						    TRUE); /* turn on gconf key */
 #else
